from pygears.conf import reg, PluginBase, Inject, inject
from pygears.core.port import InPort, OutPort, Port, HDLConsumer, HDLProducer
from pygears.core.hier_node import HierNode

sim_reg = None


class PathError(Exception):
    pass


def get_sim_map_gear(gear):
    global sim_reg
    if sim_reg is None:
        sim_reg = reg['sim']

    sim_map = sim_reg['map']
    while gear is not None:
        if gear in sim_map:
            return sim_map[gear]
        gear = gear.parent

    return None


def get_sim_cls_parent(gear_inst):
    parent = gear_inst.parent
    while parent is not None:
        if parent.params.get('sim_cls', None):
            break

        parent = parent.parent

    return parent


def is_end_consumer(port: Port, sim=False):
    if sim:
        if isinstance(port, InPort):
            if port.gear.params.get('sim_cls', None) is not None:
                return True

            if get_sim_cls_parent(port.gear):
                return False

    intf = port.consumer

    if intf is None:
        return True

    if (len(intf.consumers) == 1 and isinstance(intf.consumers[0], HDLConsumer)):
        return True

    return False


def is_source_producer(obj, sim=False):
    if not obj.consumers:
        return False

    if sim:
        # Check if intf drives output port of a sim_cls module
        for c in obj.consumers:
            if not isinstance(c, OutPort):
                continue

            if c.gear.params.get('sim_cls', None) is not None:
                return True

        for c in obj.consumers + [obj.producer]:
            if not isinstance(c, OutPort):
                continue

            if get_sim_cls_parent(c.gear):
                return False

    return isinstance(obj.producer, HDLProducer)


def get_source_producer(obj, sim=False):
    if isinstance(obj, Port):
        obj = obj.producer

    if is_source_producer(obj, sim=sim):
        return obj

    if isinstance(obj.producer, HDLProducer) or obj.producer is None:
        if sim:
            raise PathError(f'No producer found on beginning of the path: ?')

        return obj

    err = None
    try:
        return get_source_producer(obj.producer, sim=sim)
    except PathError as e:
        err = e

    raise PathError(f'{str(err)} -> {obj.name}')


def _get_consumer_tree_rec(root_intf, cur_intf, consumers, end_producer):
    for port in cur_intf.consumers:
        if isinstance(port, HDLConsumer):
            continue

        cons_intf = port.consumer
        if is_end_consumer(port, sim=True):
            end_producer[port] = (root_intf, len(consumers))
            consumers.append(port)
        elif cons_intf is not None:
            start = len(consumers)
            _get_consumer_tree_rec(root_intf, cons_intf, consumers, end_producer)
            if len(consumers) - start > 1:
                end_producer[port] = (root_intf, slice(start, len(consumers)))
            else:
                end_producer[port] = (root_intf, start)


def get_consumer_tree(intf):
    consumer_tree = reg['graph/consumer_tree']
    end_producer = reg['graph/end_producer']
    if intf in consumer_tree:
        return consumer_tree[intf]

    consumers = []
    _get_consumer_tree_rec(intf, intf, consumers, end_producer)
    return consumers


class IntfOperPlugin(PluginBase):
    @classmethod
    def bind(cls):
        global sim_reg
        sim_reg = None
        reg['graph/consumer_tree'] = {}
        reg['graph/end_producer'] = {}


def get_end_producer(obj):
    end_producer = reg['graph/end_producer']
    if obj not in end_producer:
        intf = get_source_producer(obj, sim=True)
        get_consumer_tree(intf)

    # TODO: investigate why this is necessary
    if obj not in end_producer:
        return None

    return end_producer[obj]


def get_producer_queue(obj):
    intf, i = get_end_producer(obj.producer)
    if i >= len(intf.out_queues):
        # TODO: Investigate this. This happens when consumers tries to get
        # data, but somewhere along the path from the producer to the consumer,
        # some intermediate consumer is not registered with its immediate producer
        return None

    return intf.out_queues[i]


def get_producer_port(obj):
    if isinstance(obj, Port):
        obj = obj.producer

    return obj.producer


def has_async_producer(obj):
    port = get_producer_port(obj)
    return isinstance(port, HDLProducer)



from . import ir
from .ast.visitor import Context, GearContext, FuncContext, HLSSyntaxError
from .ir_utils import Scope, HDLVisitor, is_intf_id

from . import ir_builtins

__all__ = ['Context', 'GearContext', 'FuncContext', 'Scope', 'HDLVisitor', 'ir', 'SyntaxError']



import os
from pygears.core.hier_node import HierVisitorBase
from pygears.util.fileio import save_file
from pygears import reg
from pygears.hdl import mod_lang


class HDLGenGenerateVisitor(HierVisitorBase):
    def __init__(self, outdir):
        self.hdlgen_map = reg[f'hdlgen/map']
        self.templenv = {lang: reg[f'{lang}gen/templenv'] for lang in ['sv', 'v']}

        self.outdir = outdir

    def Gear(self, node):
        lang = mod_lang(node)
        hdlgen = self.hdlgen_map.get(node, None)

        if hdlgen is not None:
            hdlgen.generate(self.templenv[lang], self.outdir)


def generate(top, conf):
    v = HDLGenGenerateVisitor(conf['outdir'])
    v.visit(top)
    return top



# TODO: generalize this to all gear generation methods. Do proper error
# reporting where it is not feasible to replace names

sv_keywords = [
    'alias',
    'always',
    'always_comb',
    'always_ff',
    'always_latch',
    'and',
    'assert',
    'assign',
    'assume',
    'automatic',
    'before',
    'begin',
    'bind',
    'bins',
    'binsof',
    'bit',
    'break',
    'buf',
    'bufif0',
    'bufif1',
    'byte',
    'case',
    'casex',
    'casez',
    'cell',
    'chandle',
    'class',
    'clocking',
    'cmos',
    'config',
    'const',
    'constraint',
    'context',
    'continue',
    'cover',
    'covergroup',
    'coverpoint',
    'cross',
    'deassign',
    'default',
    'defparam',
    'design',
    'disable',
    'dist',
    'do',
    'edge',
    'else',
    'end',
    'endcase',
    'endclass',
    'endclocking',
    'endconfig',
    'endfunction',
    'endgenerate',
    'endgroup',
    'endinterface',
    'endmodule',
    'endpackage',
    'endprimitive',
    'endprogram',
    'endproperty',
    'endspecify',
    'endsequence',
    'endtable',
    'endtask',
    'enum',
    'event',
    'expect',
    'export',
    'extends',
    'extern',
    'final',
    'first_match',
    'for',
    'force',
    'foreach',
    'forever',
    'fork',
    'forkjoin',
    'function',
    'generate',
    'genvar',
    'highz0',
    'highz1',
    'if',
    'iff',
    'ifnone',
    'ignore_bins',
    'illegal_bins',
    'import',
    'incdir',
    'include',
    'initial',
    'inout',
    'input',
    'inside',
    'instance',
    'int',
    'integer',
    'interface',
    'intersect',
    'join',
    'join_any',
    'join_none',
    'large',
    'liblist',
    'library',
    'local',
    'localparam',
    'logic',
    'longint',
    'macromodule',
    'matches',
    'medium',
    'modport',
    'module',
    'nand',
    'negedge',
    'new',
    'nmos',
    'nor',
    'noshowcancelled',
    'not',
    'notif0',
    'notif1',
    'null',
    'or',
    'output',
    'package',
    'packed',
    'parameter',
    'pmos',
    'posedge',
    'primitive',
    'priority',
    'program',
    'property',
    'protected',
    'pull0',
    'pull1',
    'pulldown',
    'pullup',
    'pulsestyle_onevent',
    'pulsestyle_ondetect',
    'pure',
    'rand',
    'randc',
    'randcase',
    'randsequence',
    'rcmos',
    'real',
    'realtime',
    'ref',
    'reg',
    'release',
    'repeat',
    'return',
    'rnmos',
    'rpmos',
    'rtran',
    'rtranif0',
    'rtranif1',
    'scalared',
    'sequence',
    'shortint',
    'shortreal',
    'showcancelled',
    'signed',
    'small',
    'solve',
    'specify',
    'specparam',
    'static',
    'string',
    'strong0',
    'strong1',
    'struct',
    'super',
    'supply0',
    'supply1',
    'table',
    'tagged',
    'task',
    'this',
    'throughout',
    'time',
    'timeprecision',
    'timeunit',
    'tran',
    'tranif0',
    'tranif1',
    'tri',
    'tri0',
    'tri1',
    'triand',
    'trior',
    'trireg',
    'type',
    'typedef',
    'union',
    'unique',
    'unsigned',
    'use',
    'uwire',
    'var',
    'vectored',
    'virtual',
    'void',
    'wait',
    'wait_order',
    'wand',
    'weak0',
    'weak1',
    'while',
    'wildcard',
    'wire',
    'with',
    'within',
    'wor',
    'xnor',
    'xor',
]



import inspect
import logging
import os

from pygears import PluginBase, reg
from pygears.conf import register_custom_log
from pygears.core.hier_node import HierVisitorBase
from .intf import SVIntfGen
from pygears.definitions import LIB_SVLIB_DIR, USER_SVLIB_DIR, USER_VLIB_DIR, LIB_VLIB_DIR
from pygears.conf import inject, Inject
from pygears.hdl import hdlmod, list_hdl_files


class SVGenInstVisitor(HierVisitorBase):
    def Gear(self, node):
        hdlgen_inst = hdlmod(node)

        if not hdlgen_inst.hierarchical:
            return True

        for i in node.local_intfs:
            hdlgen_map = reg['hdlgen/map']
            hdlgen_map[i] = SVIntfGen(i, hdlgen_inst.lang)


def svgen_inst(top, conf):
    v = SVGenInstVisitor()
    v.visit(top)

    list_hdl_files(top.name,
                   outdir=conf['outdir'],
                   rtl_only=True,
                   wrapper=conf.get('wrapper', False))

    # hdlmods = reg['hdlgen/hdlmods']
    # for fn in list_hdl_files(top, conf['outdir'], wrapper=conf.get('wrapper', False)):
    #     modname, lang = os.path.splitext(os.path.basename(fn))
    #     hdlmods[(modname, lang[1:])] = fn


def svgen_log():
    return logging.getLogger('hdlgen')


def svgen_include_get(cfg):
    return reg['hdl/include'] + [USER_SVLIB_DIR, LIB_SVLIB_DIR]


def vgen_include_get(cfg):
    return reg['hdl/include'] + [USER_VLIB_DIR, LIB_VLIB_DIR]


class SVGenInstPlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['hdlgen/map'] = {}
        reg['hdlgen/hdlmods'] = {}
        reg['hdlgen/disambig'] = {}
        register_custom_log('svgen', logging.WARNING)
        reg.confdef('svgen/include', getter=svgen_include_get)
        reg.confdef('vhdgen/include', [])

        reg['vgen/map'] = {}
        register_custom_log('vgen', logging.WARNING)
        reg.confdef('vgen/include', getter=vgen_include_get)

    @classmethod
    def reset(cls):
        reg['vgen/map'] = {}
        reg['hdlgen/hdlmods'] = {}
        reg['hdlgen/disambig'] = {}



from abc import ABC, abstractmethod
from pygears import reg
import hashlib
import functools
import typing
from pygears.hdl import mod_lang


class ResolverTypeError(Exception):
    pass


def path_name(path):
    if path.startswith('/'):
        path = path[1:]

    full_name = path.replace('/', '_')
    if len(full_name) > 100:
        path_l = path.split('/')
        head = '_'.join(path_l[:3])
        tail = '_'.join(path_l[-3:])
        mid = '_'.join(path_l[3:-3])
        full_name = head + '_' + hashlib.sha1(
            mid.encode()).hexdigest()[:8] + '_' + tail

    return full_name


class ResolverBase(ABC):
    def __init__(self, node):
        raise ResolverTypeError

    @property
    def lang(self):
        return mod_lang(self.node)

    @property
    @functools.lru_cache()
    def cfg(self):
        hdl = {}
        if 'hdl' in self.node.meta_kwds:
            hdl.update(self.node.meta_kwds['hdl'])

        if '__hdl__' in self.node.params and self.node.params['__hdl__'] is not None:
            hdl.update(self.node.params['__hdl__'])

        return hdl

    @property
    def node_def_name(self):
        if self.node.definition:
            return self.node.definition.__name__
        else:
            return self.node.name

    @property
    @functools.lru_cache()
    def hier_path_name(self):
        return path_name(self.node.name)

    @property
    @abstractmethod
    def module_name(self) -> str:
        pass

    @property
    @abstractmethod
    def file_basename(self) -> str:
        pass

    @property
    @abstractmethod
    def params(self):
        pass

    @property
    @abstractmethod
    def files(self) -> typing.List[str]:
        pass

    @abstractmethod
    def generate(self, template_env, outdir):
        pass



import collections

from pygears.conf import reg
from pygears.typing.base import GenericMeta, param_subs, is_type, T

from pygears.typing import TypeMatchError, get_match_conds


def is_type_iterable(t):
    return (not isinstance(t, (str, bytes))) and isinstance(t, collections.abc.Iterable)


def _copy_field_names(t, pat):
    if isinstance(t, GenericMeta) and isinstance(pat, GenericMeta) and (issubclass(
            t.base, pat.base)):
        for ta, pa in zip(t.args, pat.args):
            _copy_field_names(ta, pa)

        if hasattr(pat, '__parameters__'):
            t.__parameters__ = pat.__parameters__


def copy_field_names(t, pat):
    t = t.copy()
    _copy_field_names(t, pat)
    return t


def type_is_specified(t):
    if isinstance(t, T):
        return False

    if is_type(t):
        return t.specified

    if t is None:
        return True
    elif isinstance(t, dict):
        return all(type_is_specified(subt) for subt in t.values())
    elif is_type_iterable(t):
        return all(type_is_specified(subt) for subt in t)
    elif isinstance(t, bytes):
        return False
    else:
        return True


def resolve_param(val, match, namespace):
    if isinstance(val, T):
        new_p = param_subs(val, match, namespace)
        if new_p != val:
            return True, new_p
        else:
            return False, None

    is_templated_type = (isinstance(val, GenericMeta) and (not val.is_generic()))

    if ((is_templated_type or is_type_iterable(val)) and (not type_is_specified(val))):
        new_p = param_subs(val, match, namespace)

        if repr(new_p) != repr(val):
            return True, new_p

    elif isinstance(val, bytes):
        return True, eval(val, namespace, match)

    return False, None


def infer_ftypes(params, args, namespace={}):
    # Add all registered objects (types and transformations) to the namespace
    namespace = dict(namespace)
    namespace.update(reg['gear/type_arith'])

    def is_postponed(name, val):
        if isinstance(val, bytes):
            return True
        if (name in args):
            return True

        if (name == 'return'):
            return not type_is_specified(val)

        return False

    postponed = {name: val for name, val in params.items() if is_postponed(name, val)}
    match = {name: val for name, val in params.items() if name not in postponed}

    substituted = True
    final_check = False
    # Allow for keyword argument values to be templates and provide
    # a mechanism to resolve these template arguments
    while substituted or final_check:
        substituted = False
        # Loops until none of the parameters has been additionally resolved
        for name, val in postponed.copy().items():

            if name in args:
                try:
                    templ = val
                    if isinstance(val, bytes):
                        templ = templ.decode()

                    match_update, res = get_match_conds(args[name], templ, match)
                    match.update(match_update)
                    args[name] = res

                    if type_is_specified(res):
                        if is_type(res):
                            res = copy_field_names(res, params[name])

                        args[name] = res
                        match[name] = res
                        del postponed[name]
                        substituted = True
                        break
                    else:
                        postponed[name] = res

                except TypeMatchError as e:
                    err = TypeMatchError(
                        f'{str(e)}\n - when deducing type for argument '
                        f'"{name}"')
                    err.params = match
                    raise err
            else:
                try:
                    substituted, new_p = resolve_param(val, match, namespace)
                    if substituted and (name == 'return'):
                        substituted = type_is_specified(new_p)

                    if substituted:
                        if name == 'return':
                            substituted = type_is_specified(new_p)

                        match[name] = new_p
                        del postponed[name]
                        break
                    elif final_check:
                        if new_p is not None:
                            raise TypeMatchError(f'Incomplete type: {repr(new_p)}')
                        else:
                            raise TypeMatchError(f'Incomplete type: {repr(val)}')

                except Exception as e:
                    if final_check:
                        try:
                            arg_repr = str(e.args[0])
                        except:
                            arg_repr = repr(e.args[0])

                        if isinstance(val, bytes):
                            val_repr = val.decode()
                        else:
                            try:
                                val_repr = str(val)
                            except:
                                val_repr = repr(val)

                        if name == 'return':
                            err = type(e)(
                                f'{arg_repr}\n - when resolving '
                                f'return type "{val_repr}"')
                        else:
                            err = type(e)(
                                f'{arg_repr}\n - when resolving '
                                f'parameter "{name}": "{val_repr}"')
                        err.params = match
                        raise err

        final_check = not substituted and not final_check

    # print('Final postponed: ', postponed)
    # print('Final match: ', match)

    for name, val in args.items():
        get_match_conds(val, match[name], {})

    if postponed:
        name, value = next(iter(postponed.items()))
        err = TypeMatchError(f'Parameter "{name}" unresolved: {value}')
        err.params = match
        raise err

    return match



from pygears import gear
from pygears.typing import Uint
from pygears.sim import log


@gear
async def unary(din: Uint['w_data']) -> Uint['2**(int(w_data)-1)']:
    '''Returns the unary representation of a binary number'''
    async with din as val:
        if val > 2**(din.dtype.width - 1):
            log.error(
                f'{val} supplied, but only numbers <= {2**(din.dtype.width-1)} supported for this instance'
            )

        if val == 0:
            yield 0
        else:
            yield int('1' * int(val), 2)



import os
import inspect
from pygears.core.hier_node import HierYielderBase
from pygears.core.gear import Gear


class NodeYielder(HierYielderBase):
    def Gear(self, node):
        yield node

def _py_file_enum_raw(top):
    for node in NodeYielder().visit(top):
        if not isinstance(node, Gear):
            continue

        yield os.path.abspath(inspect.getfile(node.func))

        for t in node.trace:
            yield os.path.abspath(inspect.getframeinfo(t[0]).filename)


def py_files_enum(top):
    for fn in _py_file_enum_raw(top):
        if '<decorator-gen' not in fn:
            yield fn



from pygears import gear
from pygears.typing import Uint
from pygears.sim import clk


@gear
async def rom(addr: Uint, *, data, dtype, dflt=None) -> b'dtype':
    async with addr as a:
        if dflt is None:
            d = data[int(a)]
        else:
            try:
                d = data[int(a)]
            except (IndexError, KeyError):
                d = dflt

        await clk()
        yield d



import os
import logging
import sys
import time
from http.server import HTTPServer, BaseHTTPRequestHandler
from http.client import HTTPConnection
from functools import partial
from traceback import extract_tb, format_list
from pygears.conf import inject, Inject


def finish():
    os.system(f'pywave "exit"')


def pywave_running():
    return not os.system('pywave -c ""')


def pywave_reload():
    os.system(f'pywave "gtkwave::reLoadFile"')


def pywave_load_file(fn):
    if os.path.isfile(fn):
        os.system(f'pywave "gtkwave::loadFile {fn}"')


def ActivityWaveFactory(cfg):
    class ActivityWave(BaseHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super(ActivityWave, self).__init__(*args, **kwargs)

        def log_request(code='-', size='-'):
            pass

        def log_error(fmt, *args):
            pass

        def log_message(fmt, *args):
            pass

        def do_GET(self):
            """Respond to a GET request."""

            logging.info(f'Request received: {self.path}')

            try:
                self.send_response(200)
                if not self.path:
                    logging.info(f'Here?')
                    return
                elif self.path[0] == ':':
                    logging.info(f'Filedir received: {self.path}')
                    running = pywave_running()
                    if ('filedir' in cfg and cfg['filedir'] == self.path[1:]
                            and running):
                        pywave_reload()
                    else:
                        cfg['filedir'] = self.path[1:]
                        if running:
                            finish()

                        pywave_load_file(f"{cfg['filedir']}/pygears.vcd")
                        pywave_load_file(f"{cfg['filedir']}/issue.sav")

                elif self.path[0] == '/':
                    gear = self.path[1:]
                    if 'filedir' in cfg:
                        logging.info(f'Gear received: {self.path}')
                        pywave_load_file(f"{cfg['filedir']}/{gear}.sav")

                    logging.info(f'Here 0')
                    self.send_header("Content-type", "text/html")
                    self.end_headers()
                    self.wfile.write(
                        f"<html><head><title>{gear}.</title></head>".encode())

                    if 'filedir' in cfg:
                        self.wfile.write(
                            f"<body><p>Reading from: {cfg['filedir']}.</p>".
                            encode())

                    self.wfile.write(b"<p>")

                    fn = f"{cfg['filedir']}/{gear}.txt"
                    if os.path.isfile(fn):
                        with open(fn, 'r') as f:
                            for line in f:
                                self.wfile.write(f'{line}<br/>'.encode())

                    self.wfile.write(b"</p></body></html>")

            except Exception as e:
                type, value, tr = sys.exc_info()
                for s in format_list(extract_tb(tr)):
                    logging.info(s)

                logging.info(f'Error: {e}')

                raise e

    return ActivityWave


def restart_wave(sim, outdir, address):

    # conn = HTTPConnection(f'http://{address[0]}:{address[1]}')
    # conn.request('HEAD', f'http://{address[0]}:{address[1]}:{outdir}')
    # return

    while True:
        try:
            time.sleep(1)
            conn = HTTPConnection(f'{address[0]}:{address[1]}')
            conn.request('GET', f':{outdir}')
            return
        except ConnectionRefusedError:
            # print("Fail!")
            # raise e
            # return
            conn.close()
            command = [
                'python',
                os.path.join(os.path.dirname(__file__), 'activity_wave.py')
            ]

            pid = os.fork()
            if pid == 0:
                os.setsid()
                os.umask(0)
                os.execv(sys.executable, command)
            else:
                pass


@inject
def activity_wave(top,
                  sim=Inject('sim/simulator'),
                  outdir=Inject('results-dir')):
    outdir = os.path.abspath(outdir)
    sim.events['at_exit'].append(
        partial(restart_wave, outdir=outdir, address=('localhost', 5000)))


def run(address=('', 5000), server_class=HTTPServer):
    cfg = {'address': address}

    logging.basicConfig(
        filename=os.path.expanduser(f'~/activity_wave_{address[1]}.log'),
        level=getattr(logging, 'INFO'),
        format='%(asctime)s %(levelname)-8s %(message)s')

    logging.info(f'ActivityWave process started, pid={os.getpid()}')
    httpd = server_class(address, ActivityWaveFactory(cfg))
    httpd.serve_forever()


if __name__ == "__main__":
    run()



import pprint
from ..typing import Tuple


def pprint_Tuple(printer, object, stream, indent, allowance, context, level):
    if printer._compact:
        if not hasattr(object, '__parameters__'):
            printer._format(
                tuple(object.args), stream, indent, allowance, context,
                level - 1)
        else:
            printer._format(
                dict(zip(object.fields, object.args)), stream, indent,
                allowance, context, level - 1)

    else:
        stream.write('Tuple')

        cur_indent = ((level - 1) * printer._indent_per_level)
        sub_indent = (level * printer._indent_per_level)

        if object.args:
            stream.write('[')
            if not hasattr(object, '__parameters__'):
                stream.write('\n' + ' ' * (cur_indent + 1))
                repr_obj = object.args
                printer._format_items(repr_obj, stream, cur_indent,
                                      allowance + 2, context, level)
                stream.write('\n' + ' ' * cur_indent)
            else:
                stream.write('{\n' + ' ' * sub_indent)
                repr_obj = list(zip(object.fields, object.args))
                printer._format_dict_items(repr_obj, stream, cur_indent,
                                           allowance + 2, context, level)
                stream.write('\n' + ' ' * cur_indent + '}')
            stream.write(']')


pprint.PrettyPrinter._dispatch[type(Tuple).__repr__] = pprint_Tuple



import atexit

from pygears.conf import inject, Inject


class SimExtend:
    @inject
    def __init__(self, sim=Inject('sim/simulator')):
        self.sim = sim
        self.activate()

    def activate(self):
        for name, event in self.sim.events.items():
            try:
                event.append(getattr(self, name))
            except AttributeError:
                pass

        try:
            atexit.register(self.at_exit, sim=None)
        except AttributeError:
            pass

    def deactivate(self, cur_event=None):
        for name, event in self.sim.events.items():
            if name == cur_event:
                continue

            try:
                event.remove(getattr(self, name))
            except AttributeError:
                pass

        try:
            atexit.unregister(self.at_exit, sim=None)
        except AttributeError:
            pass



import cProfile
# from pygears.sim import sim
from pygears.sim.extens.sim_extend import SimExtend


class Profiler(SimExtend):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def before_run(self, sim):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def after_run(self, sim):
        self.pr.disable()
        # s = io.StringIO()
        # ps = pstats.Stats(self.pr, stream=s).sort_stats('time')
        # ps.print_stats()
        self.pr.dump_stats(self.fn)



import inspect
from ..ir_utils import IrRewriter, ir, res_true, IrVisitor, IrExprRewriter
from ..ast import Context
from ..ast.inline import call_gear
from ..ast.call import const_func_args, resolve_gear_call
from ..ast.stmt import assign_targets
from ..ast.utils import add_to_list
from .inline_cfg import Inliner
from pygears.util.utils import quiter
from pygears.lib.rng import qrange as qrange_gear


class BlockDetect(IrVisitor):
    def __init__(self):
        super().__init__()
        self.blocking = False

    def visit(self, node):
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                getattr(self, base_class.__name__)(node)
                return
        else:
            self.generic_visit(node)

    def Await(self, stmt: ir.Await):
        self.blocking = True


def is_blocking(node):
    v = BlockDetect()
    v.visit(node)

    return v.blocking


class Ununfoldable(Exception):
    pass


def exhaust(func, args):
    try:
        return list(func(*args))
    except TypeError:
        raise Ununfoldable()

    # async def exhaust(aiter):
    #     return [i async for i in aiter]

    # loop = asyncio.new_event_loop()
    # reg['gear/exec_context'] = 'sim'
    # items = loop.run_until_complete(
    #     exhaust(func_call.func.__wrapped__(*args)))
    # reg['gear/exec_context'] = 'hls'


class Unfolder(IrRewriter):
    def __init__(self, ctx):
        super().__init__()
        self.ctx = ctx
        self.generators = {}
        self.forwarded = {}

    def Expr(self, node):
        new_node = Inliner(self.forwarded, self.ctx, missing_ok=True).visit(node)
        if new_node is None:
            return node

        return new_node

    def AssignValue(self, node):
        if not isinstance(node.val, ir.GenNext):
            return super().AssignValue(node)

        gen_id = node.val.val
        if gen_id not in self.generators:
            func_call = self.ctx.scope[gen_id].func

            if not const_func_args(func_call.args.values(), func_call.kwds):
                raise Ununfoldable()

            args = tuple(node.val for node in func_call.args.values())

            vals = exhaust(func_call.func, args)

            self.generators[gen_id] = {'vals': quiter(vals)}

        next_val, last = next(self.generators[gen_id]['vals'])

        self.forwarded[node.target.name] = ir.ResExpr(next_val)
        self.generators[gen_id]['last'] = ir.ResExpr(last)

        return None

    def ExprStatement(self, node):
        if not isinstance(node.expr, (ir.GenAck, ir.GenInit)):
            return super().ExprStatement(node)

        if node.expr.val in self.generators:
            return None

        return node


class TestRewriter(IrExprRewriter):
    def __init__(self, generators):
        self.generators = generators

    def visit_GenDone(self, expr):
        if expr.val in self.generators:
            return self.generators[expr.val]['last']

        return expr


def unfold_loop(node: ir.LoopBlock, ctx: Context):
    stmts = []
    unfolder = Unfolder(ctx)
    while True:
        unfolded_loop = unfolder.visit(node)
        add_to_list(stmts, unfolded_loop.stmts)

        unfolded_test = TestRewriter(unfolder.generators).visit(node.test_loop)
        if unfolded_test == res_true:
            break

    return stmts


builtins = {range: lambda *args, **kwds: resolve_gear_call(qrange_gear.func, args, kwds)}


class HandleGenerators(IrRewriter):
    def __init__(self, ctx):
        super().__init__()
        self.ctx = ctx
        self.generators = {}

    def AssignValue(self, node):
        if not isinstance(node.val, ir.GenNext):
            return node

        gen_id = node.val.val
        func_call = self.ctx.scope[gen_id].func
        pass_eot = func_call.pass_eot

        if func_call.func in builtins:
            func_call = builtins[func_call.func](*func_call.args.values(), **func_call.kwds)

        # intf, nodes = call_gear(func_call.func, list(func_call.args.values()),
        intf, nodes = call_gear(func_call.func, func_call.args, func_call.kwds, self.ctx)

        eot_name = self.ctx.find_unique_name('_eot')
        data_name = self.ctx.find_unique_name('_data')

        intf_type = intf.dtype.dtype
        self.ctx.scope[eot_name] = ir.Variable(eot_name, intf_type.eot)
        self.ctx.scope[data_name] = ir.Variable(data_name, intf_type.data)

        # intf = intf.obj.val

        self.generators[gen_id] = {
            'intf': intf,
            'eot_name': eot_name,
            'eot': self.ctx.scope[eot_name]
        }

        if pass_eot:
            dout = ir.Component(intf, 'data')
        else:
            dout = ir.SubscriptExpr(ir.Component(intf, 'data'), ir.ResExpr(0))

        eot_load = ir.AssignValue(self.ctx.ref(eot_name),
                                  ir.SubscriptExpr(ir.Component(intf, 'data'), ir.ResExpr(-1)))

        # data_load = ir.AssignValue(
        #     self.ctx.ref(data_name),
        #     ir.Await(ir.Component(intf, 'data'),
        #              in_await=ir.Component(intf, 'valid')))

        data_load = ir.AssignValue(self.ctx.ref(data_name), dout)

        stmts = nodes + [ir.Await(ir.Component(intf, 'valid')), eot_load, data_load]

        add_to_list(stmts, assign_targets(self.ctx, node.target, dout))

        return stmts

    def ExprStatement(self, node):
        if not isinstance(node.expr, (ir.GenAck, ir.GenInit)):
            return node

        if isinstance(node.expr, ir.GenAck):
            gen_cfg = self.generators[node.expr.val]
            return ir.AssignValue(ir.Component(gen_cfg['intf'], 'ready'), ir.res_true)

        if isinstance(node.expr, ir.GenInit):
            return node

    def LoopBlock(self, node):
        # TODO: Checking for blocking coupled with blocking generator resolution. Refactor it out
        node.blocking = is_blocking(node)
        if not node.blocking:
            return node

        # try:
        #     if not is_blocking(node):
        #         return unfold_loop(node, self.ctx)
        # except Ununfoldable:
        #     pass

        node = super().LoopBlock(node)

        if not isinstance(node.test, ir.GenDone):
            return node

        gen_cfg = self.generators[node.test.val]

        eot_test = ir.BinOpExpr(
            (self.ctx.ref(gen_cfg['eot_name']), ir.ResExpr(gen_cfg['intf'].dtype.dtype.eot.max)),
            ir.opc.NotEq)

        eot_entry = ir.AssignValue(self.ctx.ref(gen_cfg['eot_name']),
                                   ir.ResExpr(gen_cfg['eot'].dtype.min))

        node.test = eot_test

        return [eot_entry, node]


def handle_generators(modblock, ctx):
    modblock = HandleGenerators(ctx).visit(modblock)

    return modblock



from .number import Number
from .base import GenericMeta


class FloatType(GenericMeta):
    @property
    def specified(self):
        return True

    def __gt__(self, others):
        return self

    def __ge__(self, others):
        return self

    def __invert__(self):
        return self

    def __neg__(self):
        return self

    def __or__(self, others):
        return self

    def __xor__(self, others):
        return self

    def __lshift__(self, others):
        return self

    def __rshift__(self, others):
        return self

    def __add__(self, other):
        return self

    __radd__ = __add__

    def __sub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __mod__(self, other):
        return other

    def __rmod__(self, other):
        return self

    __rmul__ = __mul__


class Float(float, metaclass=FloatType):
    pass


Number.register(Float)
Number.register(float)



from pygears import gear, module
from pygears.sim.extens.randomization import randomize, rand_seq
from .verif import drv


@gear
def drvrnd(*, t, cnt=None, cons=None, params=None):
    return drv(
        t=t, seq=randomize(t, module().basename, cnt=cnt, cons=cons, params=params))



import itertools

from pygears import reg
from pygears.lib.sieve import sieve
from functools import partial
from pygears.hdl.util import HDLGearHierVisitor, is_gear_instance, flow_visitor


def index_to_sv_slice(dtype, key):
    subtype = dtype[key]

    if isinstance(key, slice):
        key = min(key.start, key.stop)

    if key is None or key == 0:
        low_pos = 0
    else:
        low_pos = dtype[:key].width

    high_pos = low_pos + subtype.width - 1

    return f'{high_pos}:{low_pos}'


def get_sieve_stages_iter(node):
    for s in itertools.chain(getattr(node, 'pre_sieves', []), [node]):
        indexes = s.params['key']
        if not isinstance(indexes, tuple):
            indexes = (indexes, )

        dtype = s.in_ports[0].dtype
        out_type = s.out_ports[0].dtype
        slices = list(
            map(partial(index_to_sv_slice, dtype),
                filter(lambda i: getattr(dtype[i], 'width', 0) > 0, indexes)))
        yield slices, out_type


def get_sieve_stages(node):
    stages = list(get_sieve_stages_iter(node))
    # If any of the sieves has shrunk data to 0 width, there is nothing to
    # do
    if any(i[0] == [] for i in stages):
        stages = []

    return stages


@flow_visitor
class RemoveEqualReprSieveVisitor(HDLGearHierVisitor):
    def sieve(self, node):
        pout = node.out_ports[0]
        pin = node.in_ports[0]

        if pin.dtype == pout.dtype:
            node.bypass()


@flow_visitor
class CollapseSievesVisitor(HDLGearHierVisitor):
    def sieve(self, node):
        if not hasattr(node, 'pre_sieves'):
            node.pre_sieves = []
            node.params['stages'] = get_sieve_stages(node)

        if reg['gear/memoize']:
            return

        sieve_cons = [
            p for p in node.consumers if is_gear_instance(p.node, sieve)
        ]
        pin = node.in_ports[0]
        pout = node.out_ports[0]
        iin = pin.producer
        iout = pout.consumer

        if sieve_cons:
            # There is a Sieve connected to this Sieve, hence we can combine
            # two of them into a single SV module

            # Connect the consumers of this Sieve, which are Sieves themselves,
            # to this Sieve's predecessor
            for cons_pin in iout.consumers.copy():
                consumer = cons_pin.node
                if is_gear_instance(consumer, sieve):
                    # print(f'Merging {node.name} to {consumer.name}')
                    # print(consumer.params['key'])
                    # If the consumer is a Sieve, just register this Sieve with
                    # it, and short circuit this one
                    consumer.pre_sieves = node.pre_sieves + [node]
                    consumer.params['stages'] = get_sieve_stages(consumer)
                    iout.disconnect(cons_pin)
                    iin.connect(cons_pin)

            # print(f'Remaining conusmer: {[p.node.name for p in node.consumers]}')

            if not node.consumers:
                # Finally, if ther are no consumers left for this sieve remove
                # this Sieve completely (with all it's connections) from the
                # SVGen tree
                node.remove()
                iout.remove()



from pygears import gear
from pygears.util.utils import qrange
from pygears.typing import Queue, Tuple, Uint, Any, Bool, typeof


@gear
async def replicate(din: Tuple[{
        'length': Uint,
        'val': Any
}]) -> Queue['din["val"]']:
    """Replicates the input data. The ``length`` field of the :class:`Tuple`
    input type specifies the number of times the ``value`` field needs to be
    reproduced.

    Returns:
        A :class:`Queue` type where each element is equal to the ``value``
           input field and the `eot` signalizes the last replicated element.
    """
    async with din as (length, value):
        for i, last in qrange(length):
            yield (value, last)


@gear
async def replicate_while(cond: Bool, data: Any) -> Queue['data']:
    async with data as d:
        last = False
        while not last:
            async with cond as en:
                last = not en
                if typeof(data.dtype, Queue):
                    yield (d[0], d[1] @ last)
                else:
                    yield (d, last)


@gear
def replicate_until(cond: Bool, data: Any):
    return replicate_while(~cond, data)
